/*
 * Copyright (c) 2006-2007 Chris Smith, Shane Mc Cormack, Gregory Holmes
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package com.dmdirc.commandparser.commands.global;

import com.dmdirc.commandparser.CommandManager;
import com.dmdirc.commandparser.commands.GlobalCommand;
import com.dmdirc.commandparser.commands.IntelligentCommand;
import com.dmdirc.config.ConfigManager;
import com.dmdirc.config.Identity;
import com.dmdirc.config.IdentityManager;
import com.dmdirc.logger.ErrorLevel;
import com.dmdirc.logger.Logger;
import com.dmdirc.ui.input.AdditionalTabTargets;
import com.dmdirc.ui.interfaces.InputWindow;
import com.dmdirc.updater.UpdateChecker;
import java.io.Serializable;
import java.util.Comparator;

import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeSet;

/**
 * Provides various handy ways to test or debug the client.
 *
 * @author Chris
 */
public class Debug extends GlobalCommand implements IntelligentCommand {
    
    /**
     * Creates a new instance of Debug.
     */
    public Debug() {
        CommandManager.registerCommand(this);
    }
    
    /** {@inheritDoc} */
    public void execute(final InputWindow origin, final boolean isSilent, final String ... args) {
        if (args.length == 0) {
            showUsage(origin, isSilent, "debug", "<debug command> [options]");
        } else if ("error".equals(args[0])) {
            doError(args);
        } else if ("fakeupdate".equals(args[0])) {
            doFakeUpdate();
        } else if ("showraw".equals(args[0])) {
            doShowRaw(origin, isSilent);
        } else if ("configstats".equals(args[0])) {
            doConfigStats(origin, isSilent);
        } else if ("configinfo".equals(args[0])) {
            doConfigInfo(origin, isSilent);
        } else if ("globalconfiginfo".equals(args[0])) {
            doGlobalConfigInfo(origin, isSilent);
        } else if ("colourspam".equals(args[0])) {
            doColourSpam(origin, isSilent);
        } else if ("meminfo".equals(args[0])) {
            doMemInfo(origin, isSilent);
        } else if ("rungc".equals(args[0])) {
            doGarbage(origin, isSilent);
        } else if ("forceupdate".equals(args[0])) {
            doForceUpdate();
        } else {
            sendLine(origin, isSilent, FORMAT_ERROR, "Unknown debug action.");
        }
    }
    
    /**
     * Generates a fake error.
     *
     * @param args The arguments that were passed to the command
     */
    private void doError(final String ... args) {
        ErrorLevel el = ErrorLevel.HIGH;
        if (args.length > 2) {
            String level = args[2];
            
            if (level.equals("low")) {
                el = ErrorLevel.LOW;
            } else if (level.equals("medium")) {
                el = ErrorLevel.MEDIUM;
            } else if (level.equals("fatal")) {
                el = ErrorLevel.FATAL;
            } else if (level.equals("unknown")) {
                el = ErrorLevel.UNKNOWN;
            }
        }
        
        if (args.length > 1 && args[1].equals("user")) {
            Logger.userError(el, "Debug error message");
        } else {
            Logger.appError(el, "Debug error message", new Exception());
        }
    }
    
    /**
     * Fakes an available update.
     */
    private void doFakeUpdate() {
        new UpdateChecker().doUpdateAvailable("outofdate dummy 1337 0 http://www.example.com/");
    }
    
    /**
     * Attempts to show the server's raw window.
     *
     * @param origin The window this command was executed in
     * @param isSilent Whether this command has been silenced or not
     */
    private void doShowRaw(final InputWindow origin, final boolean isSilent) {
        if (origin == null || origin.getContainer() == null
                || origin.getContainer().getServer() == null) {
            sendLine(origin, isSilent, FORMAT_ERROR, "Cannot show raw window here.");
        } else {
            origin.getContainer().getServer().addRaw();
        }
    }
    
    /**
     * Shows stats related to the config system.
     *
     * @param origin The window this command was executed in
     * @param isSilent Whether this command has been silenced or not
     */
    private void doConfigStats(final InputWindow origin, final boolean isSilent) {
        final TreeSet<Entry<String, Integer>> sortedStats =
                new TreeSet<Entry<String, Integer>>(new ValueComparator());
        sortedStats.addAll(ConfigManager.getStats().entrySet());
        for (Map.Entry<String, Integer> entry : sortedStats) {
            sendLine(origin, isSilent, FORMAT_OUTPUT,
                    entry.getKey() + " - " + entry.getValue());
        }
    }
    
    /**
     * Shows memory usage information.
     *
     * @param origin The window this command was executed in
     * @param isSilent Whether this command has been silenced or not
     */
    private void doMemInfo(final InputWindow origin, final boolean isSilent) {
        sendLine(origin, isSilent, FORMAT_OUTPUT, "Total Memory: "
                + Runtime.getRuntime().totalMemory());
        sendLine(origin, isSilent, FORMAT_OUTPUT, "Free Memory: "
                + Runtime.getRuntime().freeMemory());
        sendLine(origin, isSilent, FORMAT_OUTPUT, "Used Memory: "
                + (Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory()));
    }
    
    /**
     * Outputs 100 lines containing various colours.
     *
     * @param origin The window this command was executed in
     * @param isSilent Whether this command has been silenced or not
     */
    private void doColourSpam(final InputWindow origin, final boolean isSilent) {
        for (int i = 0; i < 100; i++) {
            sendLine(origin, isSilent, FORMAT_OUTPUT, ((char) 3) + "5Colour! "
                    + ((char) 3) + "6Colour! " + ((char) 3) + "7Colour! "
                    + ((char) 3) + "6Colour! " + ((char) 3) + "7Colour! "
                    + ((char) 3) + "6Colour! " + ((char) 3) + "7Colour! "
                    + ((char) 3) + "6Colour! " + ((char) 3) + "7Colour! ");
        }
    }
    
    /**
     * Manually runs the garbage collector.
     *
     * @param origin The window this command was executed in
     * @param isSilent Whether this command has been silenced or not
     */
    private void doGarbage(final InputWindow origin, final boolean isSilent) {
        System.gc();
        sendLine(origin, isSilent, FORMAT_OUTPUT, "Invoked garbage collector.");
    }
    
    /**
     * Shows information about the config manager.
     *
     * @param origin The window this command was executed in
     * @param isSilent Whether this command has been silenced or not
     */
    private void doConfigInfo(final InputWindow origin, final boolean isSilent) {
        for (Identity source : origin.getConfigManager().getSources()) {
            sendLine(origin, isSilent, FORMAT_OUTPUT, source.getTarget() + " - " + source);
        }
    }
    
    /**
     * Shows information about the global config manager.
     *
     * @param origin The window this command was executed in
     * @param isSilent Whether this command has been silenced or not
     */
    private void doGlobalConfigInfo(final InputWindow origin, final boolean isSilent) {
        for (Identity source : IdentityManager.getGlobalConfig().getSources()) {
            sendLine(origin, isSilent, FORMAT_OUTPUT, source.getTarget() + " - " + source);
        }
    }
    
    /**
     * Forces the update checker to check for updates.
     */
    private void doForceUpdate() {
        new Thread(new UpdateChecker(), "Forced update checker").start();
    }
    
    /** {@inheritDoc} */
    public String getName() {
        return "debug";
    }
    
    /** {@inheritDoc} */
    public boolean showInHelp() {
        return false;
    }
    
    /** {@inheritDoc} */
    public boolean isPolyadic() {
        return true;
    }
    
    /** {@inheritDoc} */
    public int getArity() {
        return 0;
    }
    
    /** {@inheritDoc} */
    public String getHelp() {
        return null;
    }
    
    /** {@inheritDoc} */
    public AdditionalTabTargets getSuggestions(int arg, List<String> previousArgs) {
        final AdditionalTabTargets res = new AdditionalTabTargets();
        
        res.setIncludeNormal(false);
        
        if (arg == 0) {
            res.add("error");
            res.add("fakeupdate");
            res.add("showraw");
            res.add("colourspam");
            res.add("configstats");
            res.add("meminfo");
            res.add("rungc");
            res.add("configinfo");
            res.add("globalconfiginfo");
            res.add("forceupdate");
        } else if (arg == 1 && "error".equals(previousArgs.get(0))) {
            res.add("user");
            res.add("app");
        } else if (arg == 2 && "error".equals(previousArgs.get(0))) {
            res.add("low");
            res.add("medium");
            res.add("high");
            res.add("fatal");
            res.add("unknown");
        }
        
        return res;
    }
    
    /** Reverse value comparator for a map entry. */
    private static class ValueComparator implements
            Comparator<Entry<String, Integer>>, Serializable {
        
        /**
         * A version number for this class. It should be changed whenever the
         * class structure is changed (or anything else that would prevent
         * serialized objects being unserialized with the new class).
         */
        private static final long serialVersionUID = 1;
        
        /** Instantiates a new ValueComparator. */
        public ValueComparator() {
            super();
        }
        
        /** {@inheritDoc} */
        public int compare(final Entry<String, Integer> o1,
                final Entry<String, Integer> o2) {
            int returnValue = o1.getValue().compareTo(o2.getValue()) * -1;
            
            if (returnValue == 0) {
                returnValue = o1.getKey().compareToIgnoreCase(o2.getKey());
            }
            
            return returnValue;
        }
    }
    
}
